# -*- coding: utf-8 -*-
#
# AWL simulator - GUI hardware module configuration widget
#
# Copyright 2015 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.gui.configdialog import *
from awlsim.gui.util import *

from awlsim.core.hardware_loader import HwModLoader


class HwmodParamModel(QAbstractTableModel):
	# Signal: Emitted, if a new error appeared or an old error disappeared.
	newErrorText = Signal(str)

	def __init__(self):
		QAbstractTableModel.__init__(self)

		self.modDesc = None		# HwmodDescriptor
		self.modInterface = None	# HardwareInterface

	@property
	def __params(self):
		"""Return a list of tuples: (paramName, paramValue)"""
		if not self.modDesc:
			return []
		return sorted(dictItems(self.modDesc.getParameters()),
			      key = lambda p: p[0])

	def __getParamDesc(self, paramName):
		if self.modInterface:
			return self.modInterface.getParamDesc(paramName)
		return None

	def __verifyParams(self):
		if self.modInterface and self.modDesc:
			try:
				# Create a module instance.
				# This will raise errors on invalid parameters.
				self.modInterface(None, self.modDesc.getParameters())
			except AwlSimError as e:
				self.newErrorText.emit(str(e))
				return
		self.newErrorText.emit("")

	def setHwmod(self, modDesc):
		self.beginResetModel()

		try:
			if not modDesc:
				raise ValueError
			mod = HwModLoader.loadModule(modDesc.getModuleName())
			interface = mod.getInterface()
		except (AwlSimError, ValueError) as e:
			interface = None

		self.modInterface = interface
		self.modDesc = modDesc

		self.endResetModel()

	def deleteEntry(self, row):
		params = self.__params
		if row >= 0 and row < len(params):
			self.beginResetModel() # inefficient
			pName, pValue = params[row]
			self.modDesc.removeParameter(pName)
			self.endResetModel()
			self.__verifyParams()

	def rowCount(self, parent=QModelIndex()):
		return len(self.__params) + 1

	def columnCount(self, parent=QModelIndex()):
		return 2

	def data(self, index, role=Qt.DisplayRole):
		if not index:
			return None
		row, column = index.row(), index.column()
		if role in (Qt.DisplayRole, Qt.EditRole):
			params = self.__params
			if row >= len(params):
				return None
			if column == 0:
				return params[row][0]
			elif column == 1:
				paramDesc = self.__getParamDesc(params[row][0])
				if paramDesc and not paramDesc.userEditable:
					# Not user editable
					if params[row][1] is None:
						return "<set by system>"
				if params[row][1] is None:
					return ""
				return params[row][1]
		elif role == Qt.BackgroundRole:
			params = self.__params
			if row < len(params):
				paramDesc = self.__getParamDesc(params[row][0])
				if paramDesc and not paramDesc.userEditable:
					# Not user editable
					return QBrush(QColor("darkgrey"))
				if column == 0:
					if paramDesc:
						# This is a standard parameter.
						return QBrush(QColor("lightgrey"))
			return QBrush(QColor("white"))
		elif role in (Qt.ToolTipRole, Qt.WhatsThisRole):
			params = self.__params
			if row < len(params):
				paramDesc = self.__getParamDesc(params[row][0])
				if column == 0:
					if paramDesc and paramDesc.description:
						return paramDesc.description
					return "The parameter's name"
				elif column == 1:
					return "Value for '%s'" %\
						params[row][0]
			else:
				if column == 0:
					return "New parameter name"
		return None

	def headerData(self, section, orientation, role=Qt.DisplayRole):
		if role != Qt.DisplayRole:
			return None
		if orientation == Qt.Horizontal:
			return ("Parameter", "Value")[section]
		else:
			params = self.__params
			if section >= len(params):
				return "new"
			return "%d" % (section + 1)

	def setData(self, index, value, role=Qt.EditRole):
		if not index:
			return False
		row, column = index.row(), index.column()
		if role == Qt.EditRole:
			if not self.modDesc:
				return False
			params = self.__params
			if row >= len(params):
				if column == 0:
					# Add parameter
					value = value.strip()
					for pName, pValue in params:
						if pName == value:
							# Parameter does already exist
							return False
					self.beginResetModel() # inefficient
					self.modDesc.addParameter(value, "")
					self.endResetModel()
					self.__verifyParams()
					return True
			else:
				paramName = params[row][0]
				if column == 0:
					# Rename parameter
					for pName, pValue in params:
						if pName == value:
							# Parameter does already exist
							return False
					self.beginResetModel() # inefficient
					savedValue = params[row][1]
					self.modDesc.removeParameter(paramName)
					self.modDesc.addParameter(value, savedValue)
					self.endResetModel()
					self.__verifyParams()
					return True
				elif column == 1:
					# Set parameter value
					self.modDesc.setParameterValue(paramName, value)
					self.__verifyParams()
					return True
		return False

	def flags(self, index):
		if not index:
			return Qt.ItemIsEnabled
		row, column = index.row(), index.column()
		params = self.__params
		if row < len(params):
			paramDesc = self.__getParamDesc(params[row][0])
			if paramDesc and not paramDesc.userEditable:
				return Qt.ItemIsEnabled | Qt.ItemIsSelectable
		else:
			if column != 0:
				return Qt.ItemIsEnabled
		return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsEditable

class HwmodParamView(QTableView):
	def __init__(self, model=None, parent=None):
		QTableView.__init__(self, parent)

		if not model:
			model = HwmodParamModel()
		self.setModel(model)

		self.setColumnWidth(0, 170)
		self.setColumnWidth(1, 80)

	def deleteEntry(self, index=None):
		if not index:
			index = self.currentIndex()
		if not index:
			return
		self.model().deleteEntry(index.row())

	def keyPressEvent(self, ev):
		QTableView.keyPressEvent(self, ev)

		if ev.key() == Qt.Key_Delete:
			self.deleteEntry()

	def setHwmod(self, modDesc):
		self.model().setHwmod(modDesc)

class HwmodConfigWidget(QWidget):
	def __init__(self, parent=None):
		QWidget.__init__(self, parent)
		self.setLayout(QGridLayout())
		self.layout().setContentsMargins(QMargins())

		self.__loadedModDescs = []
		self.__modSelectChangeBlocked = Blocker()

		group = QGroupBox(self)
		group.setLayout(QGridLayout())

		label = QLabel("Available modules:", self)
		group.layout().addWidget(label, 0, 0)
		self.availList = QListWidget(self)
		self.availList.setMaximumWidth(180)
		group.layout().addWidget(self.availList, 1, 0)
		self.manualModName = QLineEdit(self)
		self.manualModName.setToolTip("Name of another module to add.\n"
			"Note: Typos in the module name will result in "
			"errors on CPU startup.")
		self.manualModName.setMaximumWidth(180)
		group.layout().addWidget(self.manualModName, 2, 0)

		vbox = QVBoxLayout()
		self.addButton = QPushButton(self)
		self.addButton.setIcon(getIcon("next"))
		self.addButton.setToolTip("Add the selected module to the "
			"project\nand mark it for download to the CPU.")
		vbox.addWidget(self.addButton)
		self.delButton = QPushButton(self)
		self.delButton.setIcon(getIcon("previous"))
		self.delButton.setToolTip("Remove the selected module from "
			"the project\nand mark it for removal from the CPU.")
		vbox.addWidget(self.delButton)
		group.layout().addLayout(vbox, 0, 1, 3, 1)

		label = QLabel("Loaded modules:", self)
		group.layout().addWidget(label, 0, 2)
		self.loadedList = QListWidget(self)
		self.loadedList.setMaximumWidth(180)
		group.layout().addWidget(self.loadedList, 1, 2, 2, 1)

		self.paramViewLabel = QLabel("Module parameters:", self)
		group.layout().addWidget(self.paramViewLabel, 0, 3)
		vbox = QVBoxLayout()
		self.paramView = HwmodParamView(None, self)
		self.paramView.setMinimumWidth(300)
		vbox.addWidget(self.paramView)
		self.paramErrorText = QLabel(self)
		self.paramErrorText.setWordWrap(True)
		vbox.addWidget(self.paramErrorText)
		group.layout().addLayout(vbox, 1, 3, 2, 1)

		self.layout().addWidget(group, 0, 0)

		self.layout().setRowStretch(1, 1)

		self.__handleLoadedSelectChange(None, None)
		self.__updateAddButton()
		self.setAvailableModules(HwModLoader.builtinHwModules)

		self.availList.currentItemChanged.connect(self.__handleAvailSelectChange)
		self.manualModName.textChanged.connect(self.__handleManualModChange)
		self.loadedList.currentItemChanged.connect(self.__handleLoadedSelectChange)
		self.addButton.released.connect(self.__handleAdd)
		self.delButton.released.connect(self.__handleDel)
		self.paramView.model().newErrorText.connect(self.__handleNewErrorText)

	def __updateAddButton(self):
		self.addButton.setEnabled(self.manualModName.text().strip() != "" or
					  self.availList.currentItem() is not None)

	def __handleAvailSelectChange(self, cur, prev):
		if not self.__modSelectChangeBlocked:
			with self.__modSelectChangeBlocked:
				self.manualModName.setText("")
				self.loadedList.setCurrentItem(None)
		self.__updateAddButton()

	def __handleManualModChange(self, text):
		if not self.__modSelectChangeBlocked:
			with self.__modSelectChangeBlocked:
				self.availList.setCurrentItem(None)
				self.loadedList.setCurrentItem(None)
		self.__updateAddButton()

	def __handleLoadedSelectChange(self, cur, prev):
		if cur:
			with self.__modSelectChangeBlocked:
				self.availList.setCurrentItem(None)
				self.manualModName.setText("")
			self.paramView.setHwmod(cur.data(Qt.UserRole))
		else:
			self.paramView.setHwmod(None)
		self.paramViewLabel.setEnabled(bool(cur))
		self.paramView.setEnabled(bool(cur))
		self.delButton.setEnabled(bool(cur))
		self.__handleNewErrorText("")

	def __handleAdd(self):
		manualModName = self.manualModName.text().strip()
		if manualModName:
			modDesc = HwmodDescriptor(manualModName)
			self.manualModName.clear()
		else:
			item = self.availList.currentItem()
			if not item:
				return
			modDesc = item.data(Qt.UserRole).dup()
		item = QListWidgetItem(modDesc.getModuleName(),
				       self.loadedList)
		item.setData(Qt.UserRole, modDesc)
		self.__loadedModDescs.append(modDesc)

	def __handleDel(self):
		item = self.loadedList.currentItem()
		if not item:
			return
		item = self.loadedList.takeItem(self.loadedList.row(item))
		modDesc = item.data(Qt.UserRole)
		self.__loadedModDescs.remove(modDesc)
		self.__handleLoadedSelectChange(self.loadedList.currentItem(), None)

	def __handleNewErrorText(self, text):
		if text.strip().startswith("["):
			text = text[text.find("]") + 1 : ]
		text = text.strip()
		if text:
			text = "Warning: " + text
		self.paramErrorText.setText(text)
		if text:
			self.paramErrorText.show()
		else:
			self.paramErrorText.hide()

	def setAvailableModules(self, mods):
		self.availList.clear()
		self.availableMods = mods
		for modName in mods:
			item = QListWidgetItem(modName, self.availList)
			try:
				interface = HwModLoader.loadModule(modName).getInterface()
			except AwlSimError as e:
				interface = None
			params = {}
			if interface:
				for paramDesc in interface.getParamDescs():
					if paramDesc.defaultValue is None:
						params[paramDesc.name] = None
						continue
					params[paramDesc.name] = str(paramDesc.defaultValue)
			modDesc = HwmodDescriptor(moduleName = modName,
						  parameters = params)
			item.setData(Qt.UserRole, modDesc)

	def loadFromProject(self, project):
		self.loadedList.clear()
		hwSettings = project.getHwmodSettings()
		self.__loadedModDescs = []
		for modDesc in hwSettings.getLoadedModules():
			modDesc = modDesc.dup()
			item = QListWidgetItem(modDesc.getModuleName())
			item.setData(Qt.UserRole, modDesc)
			self.loadedList.addItem(item)
			self.__loadedModDescs.append(modDesc)

	def storeToProject(self, project):
		hwSettings = project.getHwmodSettings()
		hwSettings.setLoadedModules(self.__loadedModDescs)
		return True

class HwmodConfigDialog(AbstractConfigDialog):
	def __init__(self, project, parent=None):
		AbstractConfigDialog.__init__(self,
			project = project,
			iconName = "hwmod",
			title = "Hardware module setup",
			centralWidget = HwmodConfigWidget(),
			parent = parent)

	def loadFromProject(self):
		self.centralWidget.loadFromProject(self.project)

	def storeToProject(self):
		if self.centralWidget.storeToProject(self.project):
			self.settingsChanged.emit()



# -*- coding: utf-8 -*-
#
# AWL simulator - FUP - Wire classes
#
# Copyright 2016-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.xmlfactory import *

from awlsim.gui.geo2d import *
from awlsim.gui.util import *
from awlsim.gui.fup.fup_base import *


class FupWire_factory(XmlFactory):
	def parser_open(self, tag=None):
		self.inWire = False
		XmlFactory.parser_open(self, tag)

	def parser_beginTag(self, tag):
		if not self.inWire:
			if tag.name == "wire":
				self.inWire = True
				idNum = tag.getAttrInt("id")
				if idNum in (w.idNum for w in self.grid.wires):
					raise self.Error("<wire id=%d> does "
						"already exist." % idNum)
				# Create wire and add it to the grid.
				FupWire(grid=self.grid, idNum=idNum)
				return
		XmlFactory.parser_beginTag(self, tag)

	def parser_endTag(self, tag):
		if self.inWire:
			if tag.name == "wire":
				self.inWire = False
				return
		else:
			if tag.name == "wires":
				self.parser_finish()
				return
		XmlFactory.parser_endTag(self, tag)

	def composer_getTags(self):
		return [
			self.Tag(name="wire",
				attrs={
					"id" : str(self.wire.idNum),
				}),
		]

class FupWire(FupBaseClass):
	"""FUP/FBD wire connecting two FupConn connections."""

	factory = FupWire_factory

	BRANCH_DIA = 4

	def __init__(self, grid, idNum=None):
		FupBaseClass.__init__(self)
		self.grid = grid
		self.connections = set()	# The connections this wire is connected to
		self.outConn = None		# The out-connection this is connected to

		if idNum is None:
			idNum = grid.getUnusedWireIdNum()
		self.idNum = idNum		# The ID number of this wire
		grid.addWire(self)

		self.__wirePen = QPen(QColor("#000000"))
		self.__wirePen.setWidth(2)
		self.__wireCollidingPen = QPen(QColor("#C02020"))
		self.__wireCollidingPen.setWidth(2)
		self.__wireBranchPen = QPen(QColor("#000000"))
		self.__wireBranchPen.setWidth(1)
		self.__wireBranchBrush = QBrush(QColor("#000000"))

		self.checkCollisions()

	def checkCollisions(self):
		"""Mark the wire as must-check-collisions.
		The collision check will be done at the next re-draw.
		"""
		self.__checkCollisions = True
		self.__hasCollision = False

	def connect(self, conn):
		"""Add a connection to this wire.
		"""
		if conn in self.connections:
			return
		if conn.OUT and\
		   self.outConn is not None and\
		   self.outConn is not conn:
			# We already have an output connection.
			raise ValueError
		self.connections.add(conn)
		conn.wire = self
		if conn.OUT:
			self.outConn = conn
		self.checkCollisions()

	def disconnectAll(self):
		"""Disconenct all connections.
		"""
		for conn in self.connections:
			conn.wire = None
		self.connections.clear()
		self.outConn = None
		self.grid.removeWire(self)
		self.checkCollisions()

	def disconnect(self, conn):
		"""Disconnect a connection from this wire.
		"""
		conn.wire = None
		self.connections.remove(conn)
		if self.outConn is conn:
			# Only inputs left. Remove them all.
			self.disconnectAll()
		if len(self.connections) == 1:
			# Only one connection left. Remove that, too.
			self.disconnectAll()
		if not self.connections and not self.outConn:
			self.grid.removeWire(self)
		self.checkCollisions()

	class DrawInfo(object):
		usesDirect = False

		def __init__(self, segments, segDirect):
			self.segments = segments	# Regular segments (list)
			self.segDirect = segDirect	# Direct connection segment

		@property
		def allRegularSegments(self):
			return self.segments

	class StartIntersections(object):
		posCount = 0
		negCount = 0
		horizCount = 0

	def draw(self, painter):
		if self.outConn is None:
			return # Only inputs. Do not draw.
		grid = self.grid

		# Branch circles diameter
		branchR, branchD = self.BRANCH_DIA // 2, self.BRANCH_DIA
		painter.setBrush(self.__wireBranchBrush)

		# Calculate the coordinates of all wire lines.
		cellPixWidth = self.grid.cellPixWidth
		xAbs0, yAbs0 = self.outConn.pixCoords
		x = (xAbs0 // cellPixWidth) * cellPixWidth + cellPixWidth
		segStart = LineSeg2D.fromCoords(xAbs0, yAbs0, x, yAbs0)
		wireLines = [] # List of DrawInfo()s
		for inConn in self.connections:
			if inConn is self.outConn:
				continue
			assert(inConn.IN)

			# Construct line segments to draw the wire from out to in.

			xAbs1, yAbs1 = inConn.pixCoords

			seg0 = LineSeg2D.fromCoords(x, yAbs0, x, yAbs1)
			seg1 = LineSeg2D.fromCoords(x, yAbs1, xAbs1, yAbs1)
			segDirect = LineSeg2D.fromCoords(x, yAbs0, xAbs1, yAbs1)

			wireLines.append(self.DrawInfo((seg0, seg1), segDirect))

		def drawBranch(x, y):
			painter.setPen(self.__wireBranchPen)
			painter.drawEllipse(x - branchR, y - branchR,
					    branchD, branchD)

		def drawSeg(seg, pen=self.__wirePen):
			painter.setPen(pen)
			grid.drawWireLine(painter, self, seg)

		# Check for wire collisions, if requested.
		# Store the result for future re-draws.
		if self.__checkCollisions:
			hasCollision = 0
			excludeWires = {self}
			for drawInfo in wireLines:
				hasCollision |= int(any(
					grid.checkWireLine(painter, excludeWires, seg)
					for seg in drawInfo.segments
				))
			self.__hasCollision = bool(hasCollision)
			self.__checkCollisions = False

		# Draw wire from output to all inputs
		drawSeg(segStart)
		for drawInfo in wireLines:
			if self.__hasCollision:
				drawSeg(drawInfo.segDirect,
					pen=self.__wireCollidingPen)
				drawInfo.usesDirect = True
			else:
				for seg in drawInfo.segments:
					drawSeg(seg)

		# Draw the branch circles
		startIntersections = self.StartIntersections()
		for drawInfo in wireLines:
			if drawInfo.usesDirect:
				continue
			for seg in drawInfo.allRegularSegments:
				intersections = {}
				def addInter(interPoint, otherSeg):
					if interPoint == segStart.pointB:
						vectY = otherSeg.vect.y
						if vectY == 0:
							startIntersections.horizCount += 1
						elif vectY > 0:
							startIntersections.posCount += 1
						else:
							startIntersections.negCount += 1
					else:
						key = (interPoint.xInt, interPoint.yInt)
						intersections[key] = intersections.setdefault(key, 0) + 1

				for otherDrawInfo in wireLines:
					if otherDrawInfo is drawInfo or\
					   otherDrawInfo.usesDirect:
						continue
					for otherSeg in otherDrawInfo.allRegularSegments:
						inter = seg.intersection(otherSeg)
						if not inter.intersects:
							continue
						addInter(inter.pointA, otherSeg)
						if inter.pointA != inter.pointB:
							addInter(inter.pointB, otherSeg)

				for (x, y), count in dictItems(intersections):
					if count > 1:
						drawBranch(x, y)
		# If there are at least two line segments starting from segStart
		# pointing into opposite directions, draw the start branch.
		count = int(bool(startIntersections.posCount)) +\
			int(bool(startIntersections.negCount)) +\
			int(bool(startIntersections.horizCount))
		if count >= 2:
			drawBranch(segStart.pointB.x, segStart.pointB.y)



# -*- coding: utf-8 -*-
#
# AWL simulator - Instruction types
#
# Copyright 2012-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.enumeration import *
from awlsim.common.util import *


__all__ = [
	"AwlInsnTypes",
]


class __AwlInsnTypesClass(object): #+cdef
	def __init__(self):
		EnumGen.start
		self.TYPE_U		= EnumGen.item	# U
		self.TYPE_UN		= EnumGen.item	# UN
		self.TYPE_O		= EnumGen.item	# O
		self.TYPE_ON		= EnumGen.item	# ON
		self.TYPE_X		= EnumGen.item	# X
		self.TYPE_XN		= EnumGen.item	# XN
		self.TYPE_UB		= EnumGen.item	# U(
		self.TYPE_UNB		= EnumGen.item	# UN(
		self.TYPE_OB		= EnumGen.item	# O(
		self.TYPE_ONB		= EnumGen.item	# ON(
		self.TYPE_XB		= EnumGen.item	# X(
		self.TYPE_XNB		= EnumGen.item	# XN(
		self.TYPE_BEND		= EnumGen.item	# )
		self.TYPE_ASSIGN	= EnumGen.item	# =
		self.TYPE_R		= EnumGen.item	# R
		self.TYPE_S		= EnumGen.item	# S
		self.TYPE_NOT		= EnumGen.item	# NOT
		self.TYPE_SET		= EnumGen.item	# SET
		self.TYPE_CLR		= EnumGen.item	# CLR
		self.TYPE_SAVE		= EnumGen.item	# SAVE
		self.TYPE_FN		= EnumGen.item	# FN
		self.TYPE_FP		= EnumGen.item	# FP
		self.TYPE_EQ_I		= EnumGen.item	# ==I
		self.TYPE_NE_I		= EnumGen.item	# <>I
		self.TYPE_GT_I		= EnumGen.item	# >I
		self.TYPE_LT_I		= EnumGen.item	# <I
		self.TYPE_GE_I		= EnumGen.item	# >=I
		self.TYPE_LE_I		= EnumGen.item	# <=I
		self.TYPE_EQ_D		= EnumGen.item	# ==D
		self.TYPE_NE_D		= EnumGen.item	# <>D
		self.TYPE_GT_D		= EnumGen.item	# >D
		self.TYPE_LT_D		= EnumGen.item	# <D
		self.TYPE_GE_D		= EnumGen.item	# >=D
		self.TYPE_LE_D		= EnumGen.item	# <=D
		self.TYPE_EQ_R		= EnumGen.item	# ==R
		self.TYPE_NE_R		= EnumGen.item	# <>R
		self.TYPE_GT_R		= EnumGen.item	# >R
		self.TYPE_LT_R		= EnumGen.item	# <R
		self.TYPE_GE_R		= EnumGen.item	# >=R
		self.TYPE_LE_R		= EnumGen.item	# <=R
		self.TYPE_BTI		= EnumGen.item	# BTI
		self.TYPE_ITB		= EnumGen.item	# ITB
		self.TYPE_BTD		= EnumGen.item	# BTD
		self.TYPE_ITD		= EnumGen.item	# ITD
		self.TYPE_DTB		= EnumGen.item	# DTB
		self.TYPE_DTR		= EnumGen.item	# DTR
		self.TYPE_INVI		= EnumGen.item	# INVI
		self.TYPE_INVD		= EnumGen.item	# INVD
		self.TYPE_NEGI		= EnumGen.item	# NEGI
		self.TYPE_NEGD		= EnumGen.item	# NEGD
		self.TYPE_NEGR		= EnumGen.item	# NEGR
		self.TYPE_TAW		= EnumGen.item	# TAW
		self.TYPE_TAD		= EnumGen.item	# TAD
		self.TYPE_RND		= EnumGen.item	# RND
		self.TYPE_TRUNC		= EnumGen.item	# TRUNC
		self.TYPE_RNDP		= EnumGen.item	# RND+
		self.TYPE_RNDN		= EnumGen.item	# RND-
		self.TYPE_FR		= EnumGen.item	# FR
		self.TYPE_L		= EnumGen.item	# L
		self.TYPE_LC		= EnumGen.item	# LC
		self.TYPE_ZV		= EnumGen.item	# ZV
		self.TYPE_ZR		= EnumGen.item	# ZR
		self.TYPE_AUF		= EnumGen.item	# AUF
		self.TYPE_TDB		= EnumGen.item	# TDB
		self.TYPE_SPA		= EnumGen.item	# SPA
		self.TYPE_SPL		= EnumGen.item	# SPL
		self.TYPE_SPB		= EnumGen.item	# SPB
		self.TYPE_SPBN		= EnumGen.item	# SPBN
		self.TYPE_SPBB		= EnumGen.item	# SPBB
		self.TYPE_SPBNB		= EnumGen.item	# SPBNB
		self.TYPE_SPBI		= EnumGen.item	# SPBI
		self.TYPE_SPBIN		= EnumGen.item	# SPBIN
		self.TYPE_SPO		= EnumGen.item	# SPO
		self.TYPE_SPS		= EnumGen.item	# SPS
		self.TYPE_SPZ		= EnumGen.item	# SPZ
		self.TYPE_SPN		= EnumGen.item	# SPN
		self.TYPE_SPP		= EnumGen.item	# SPP
		self.TYPE_SPM		= EnumGen.item	# SPM
		self.TYPE_SPPZ		= EnumGen.item	# SPPZ
		self.TYPE_SPMZ		= EnumGen.item	# SPMZ
		self.TYPE_SPU		= EnumGen.item	# SPU
		self.TYPE_LOOP		= EnumGen.item	# LOOP
		self.TYPE_PL_I		= EnumGen.item	# +I
		self.TYPE_MI_I		= EnumGen.item	# -I
		self.TYPE_MU_I		= EnumGen.item	# *I
		self.TYPE_DI_I		= EnumGen.item	# /I
		self.TYPE_PL		= EnumGen.item	# +
		self.TYPE_PL_D		= EnumGen.item	# +D
		self.TYPE_MI_D		= EnumGen.item	# -D
		self.TYPE_MU_D		= EnumGen.item	# *D
		self.TYPE_DI_D		= EnumGen.item	# /D
		self.TYPE_MOD		= EnumGen.item	# MOD
		self.TYPE_PL_R		= EnumGen.item	# +R
		self.TYPE_MI_R		= EnumGen.item	# -R
		self.TYPE_MU_R		= EnumGen.item	# *R
		self.TYPE_DI_R		= EnumGen.item	# /R
		self.TYPE_ABS		= EnumGen.item	# ABS
		self.TYPE_SQR		= EnumGen.item	# SQR
		self.TYPE_SQRT		= EnumGen.item	# SQRT
		self.TYPE_EXP		= EnumGen.item	# EXP
		self.TYPE_LN		= EnumGen.item	# LN
		self.TYPE_SIN		= EnumGen.item	# SIN
		self.TYPE_COS		= EnumGen.item	# COS
		self.TYPE_TAN		= EnumGen.item	# TAN
		self.TYPE_ASIN		= EnumGen.item	# ASIN
		self.TYPE_ACOS		= EnumGen.item	# ACOS
		self.TYPE_ATAN		= EnumGen.item	# ATAN
		self.TYPE_LAR1		= EnumGen.item	# LAR1
		self.TYPE_LAR2		= EnumGen.item	# LAR2
		self.TYPE_T		= EnumGen.item	# T
		self.TYPE_TAR		= EnumGen.item	# TAR
		self.TYPE_TAR1		= EnumGen.item	# TAR1
		self.TYPE_TAR2		= EnumGen.item	# TAR2
		self.TYPE_BE		= EnumGen.item	# BE
		self.TYPE_BEB		= EnumGen.item	# BEB
		self.TYPE_BEA		= EnumGen.item	# BEA
		self.TYPE_CALL		= EnumGen.item	# CALL
		self.TYPE_CC		= EnumGen.item	# CC
		self.TYPE_UC		= EnumGen.item	# UC
		self.TYPE_MCRB		= EnumGen.item	# MCR(
		self.TYPE_BMCR		= EnumGen.item	# )MCR
		self.TYPE_MCRA		= EnumGen.item	# MCRA
		self.TYPE_MCRD		= EnumGen.item	# MCRD
		self.TYPE_SSI		= EnumGen.item	# SSI
		self.TYPE_SSD		= EnumGen.item	# SSD
		self.TYPE_SLW		= EnumGen.item	# SLW
		self.TYPE_SRW		= EnumGen.item	# SRW
		self.TYPE_SLD		= EnumGen.item	# SLD
		self.TYPE_SRD		= EnumGen.item	# SRD
		self.TYPE_RLD		= EnumGen.item	# RLD
		self.TYPE_RRD		= EnumGen.item	# RRD
		self.TYPE_RLDA		= EnumGen.item	# RLDA
		self.TYPE_RRDA		= EnumGen.item	# RRDA
		self.TYPE_SI		= EnumGen.item	# SI
		self.TYPE_SV		= EnumGen.item	# SV
		self.TYPE_SE		= EnumGen.item	# SE
		self.TYPE_SS		= EnumGen.item	# SS
		self.TYPE_SA		= EnumGen.item	# SA
		self.TYPE_UW		= EnumGen.item	# UW
		self.TYPE_OW		= EnumGen.item	# OW
		self.TYPE_XOW		= EnumGen.item	# XOW
		self.TYPE_UD		= EnumGen.item	# UD
		self.TYPE_OD		= EnumGen.item	# OD
		self.TYPE_XOD		= EnumGen.item	# XOD
		self.TYPE_TAK		= EnumGen.item	# TAK
		self.TYPE_PUSH		= EnumGen.item	# PUSH
		self.TYPE_POP		= EnumGen.item	# POP
		self.TYPE_ENT		= EnumGen.item	# ENT
		self.TYPE_LEAVE		= EnumGen.item	# LEAVE
		self.TYPE_INC		= EnumGen.item	# INC
		self.TYPE_DEC		= EnumGen.item	# DEC
		self.TYPE_INCAR1	= EnumGen.item	# +AR1
		self.TYPE_INCAR2	= EnumGen.item	# +AR2
		self.TYPE_BLD		= EnumGen.item	# BLD
		self.TYPE_NOP		= EnumGen.item	# NOP
		# Special instructions for debugging of the simulator
		self.TYPE_EXTENDED	= EnumGen.itemNoInc
		self.TYPE_ASSERT_EQ	= EnumGen.item	# __ASSERT==
		self.TYPE_ASSERT_EQ_R	= EnumGen.item 	# __ASSERT==R
		self.TYPE_ASSERT_NE	= EnumGen.item 	# __ASSERT<>
		self.TYPE_ASSERT_GT	= EnumGen.item 	# __ASSERT>
		self.TYPE_ASSERT_LT	= EnumGen.item 	# __ASSERT<
		self.TYPE_ASSERT_GE	= EnumGen.item 	# __ASSERT>=
		self.TYPE_ASSERT_LE	= EnumGen.item 	# __ASSERT<=
		self.TYPE_SLEEP		= EnumGen.item 	# __SLEEP
		self.TYPE_STWRST	= EnumGen.item 	# __STWRST
		self.TYPE_FEATURE	= EnumGen.item 	# __FEATURE
		# Special instructions for internal usage
		self.TYPE_GENERIC_CALL	= EnumGen.item	# No mnemonic
		EnumGen.end

		self.name2type_german = {
			"U"	: self.TYPE_U,
			"UN"	: self.TYPE_UN,
			"O"	: self.TYPE_O,
			"ON"	: self.TYPE_ON,
			"X"	: self.TYPE_X,
			"XN"	: self.TYPE_XN,
			"U("	: self.TYPE_UB,
			"UN("	: self.TYPE_UNB,
			"O("	: self.TYPE_OB,
			"ON("	: self.TYPE_ONB,
			"X("	: self.TYPE_XB,
			"XN("	: self.TYPE_XNB,
			")"	: self.TYPE_BEND,
			"="	: self.TYPE_ASSIGN,
			"R"	: self.TYPE_R,
			"S"	: self.TYPE_S,
			"NOT"	: self.TYPE_NOT,
			"SET"	: self.TYPE_SET,
			"CLR"	: self.TYPE_CLR,
			"SAVE"	: self.TYPE_SAVE,
			"FN"	: self.TYPE_FN,
			"FP"	: self.TYPE_FP,
			"==I"	: self.TYPE_EQ_I,
			"<>I"	: self.TYPE_NE_I,
			">I"	: self.TYPE_GT_I,
			"<I"	: self.TYPE_LT_I,
			">=I"	: self.TYPE_GE_I,
			"<=I"	: self.TYPE_LE_I,
			"==D"	: self.TYPE_EQ_D,
			"<>D"	: self.TYPE_NE_D,
			">D"	: self.TYPE_GT_D,
			"<D"	: self.TYPE_LT_D,
			">=D"	: self.TYPE_GE_D,
			"<=D"	: self.TYPE_LE_D,
			"==R"	: self.TYPE_EQ_R,
			"<>R"	: self.TYPE_NE_R,
			">R"	: self.TYPE_GT_R,
			"<R"	: self.TYPE_LT_R,
			">=R"	: self.TYPE_GE_R,
			"<=R"	: self.TYPE_LE_R,
			"BTI"	: self.TYPE_BTI,
			"ITB"	: self.TYPE_ITB,
			"BTD"	: self.TYPE_BTD,
			"ITD"	: self.TYPE_ITD,
			"DTB"	: self.TYPE_DTB,
			"DTR"	: self.TYPE_DTR,
			"INVI"	: self.TYPE_INVI,
			"INVD"	: self.TYPE_INVD,
			"NEGI"	: self.TYPE_NEGI,
			"NEGD"	: self.TYPE_NEGD,
			"NEGR"	: self.TYPE_NEGR,
			"TAW"	: self.TYPE_TAW,
			"TAD"	: self.TYPE_TAD,
			"RND"	: self.TYPE_RND,
			"TRUNC"	: self.TYPE_TRUNC,
			"RND+"	: self.TYPE_RNDP,
			"RND-"	: self.TYPE_RNDN,
			"FR"	: self.TYPE_FR,
			"L"	: self.TYPE_L,
			"LC"	: self.TYPE_LC,
			"ZV"	: self.TYPE_ZV,
			"ZR"	: self.TYPE_ZR,
			"AUF"	: self.TYPE_AUF,
			"TDB"	: self.TYPE_TDB,
			"SPA"	: self.TYPE_SPA,
			"SPL"	: self.TYPE_SPL,
			"SPB"	: self.TYPE_SPB,
			"SPBN"	: self.TYPE_SPBN,
			"SPBB"	: self.TYPE_SPBB,
			"SPBNB"	: self.TYPE_SPBNB,
			"SPBI"	: self.TYPE_SPBI,
			"SPBIN"	: self.TYPE_SPBIN,
			"SPO"	: self.TYPE_SPO,
			"SPS"	: self.TYPE_SPS,
			"SPZ"	: self.TYPE_SPZ,
			"SPN"	: self.TYPE_SPN,
			"SPP"	: self.TYPE_SPP,
			"SPM"	: self.TYPE_SPM,
			"SPPZ"	: self.TYPE_SPPZ,
			"SPMZ"	: self.TYPE_SPMZ,
			"SPU"	: self.TYPE_SPU,
			"LOOP"	: self.TYPE_LOOP,
			"+I"	: self.TYPE_PL_I,
			"-I"	: self.TYPE_MI_I,
			"*I"	: self.TYPE_MU_I,
			"/I"	: self.TYPE_DI_I,
			"+"	: self.TYPE_PL,
			"+D"	: self.TYPE_PL_D,
			"-D"	: self.TYPE_MI_D,
			"*D"	: self.TYPE_MU_D,
			"/D"	: self.TYPE_DI_D,
			"MOD"	: self.TYPE_MOD,
			"+R"	: self.TYPE_PL_R,
			"-R"	: self.TYPE_MI_R,
			"*R"	: self.TYPE_MU_R,
			"/R"	: self.TYPE_DI_R,
			"ABS"	: self.TYPE_ABS,
			"SQR"	: self.TYPE_SQR,
			"SQRT"	: self.TYPE_SQRT,
			"EXP"	: self.TYPE_EXP,
			"LN"	: self.TYPE_LN,
			"SIN"	: self.TYPE_SIN,
			"COS"	: self.TYPE_COS,
			"TAN"	: self.TYPE_TAN,
			"ASIN"	: self.TYPE_ASIN,
			"ACOS"	: self.TYPE_ACOS,
			"ATAN"	: self.TYPE_ATAN,
			"LAR1"	: self.TYPE_LAR1,
			"LAR2"	: self.TYPE_LAR2,
			"T"	: self.TYPE_T,
			"TAR"	: self.TYPE_TAR,
			"TAR1"	: self.TYPE_TAR1,
			"TAR2"	: self.TYPE_TAR2,
			"BE"	: self.TYPE_BE,
			"BEB"	: self.TYPE_BEB,
			"BEA"	: self.TYPE_BEA,
			"CALL"	: self.TYPE_CALL,
			"CC"	: self.TYPE_CC,
			"UC"	: self.TYPE_UC,
			"MCR("	: self.TYPE_MCRB,
			")MCR"	: self.TYPE_BMCR,
			"MCRA"	: self.TYPE_MCRA,
			"MCRD"	: self.TYPE_MCRD,
			"SSI"	: self.TYPE_SSI,
			"SSD"	: self.TYPE_SSD,
			"SLW"	: self.TYPE_SLW,
			"SRW"	: self.TYPE_SRW,
			"SLD"	: self.TYPE_SLD,
			"SRD"	: self.TYPE_SRD,
			"RLD"	: self.TYPE_RLD,
			"RRD"	: self.TYPE_RRD,
			"RLDA"	: self.TYPE_RLDA,
			"RRDA"	: self.TYPE_RRDA,
			"SI"	: self.TYPE_SI,
			"SV"	: self.TYPE_SV,
			"SE"	: self.TYPE_SE,
			"SS"	: self.TYPE_SS,
			"SA"	: self.TYPE_SA,
			"UW"	: self.TYPE_UW,
			"OW"	: self.TYPE_OW,
			"XOW"	: self.TYPE_XOW,
			"UD"	: self.TYPE_UD,
			"OD"	: self.TYPE_OD,
			"XOD"	: self.TYPE_XOD,
			"TAK"	: self.TYPE_TAK,
			"PUSH"	: self.TYPE_PUSH,
			"POP"	: self.TYPE_POP,
			"ENT"	: self.TYPE_ENT,
			"LEAVE"	: self.TYPE_LEAVE,
			"INC"	: self.TYPE_INC,
			"DEC"	: self.TYPE_DEC,
			"+AR1"	: self.TYPE_INCAR1,
			"+AR2"	: self.TYPE_INCAR2,
			"BLD"	: self.TYPE_BLD,
			"NOP"	: self.TYPE_NOP,

			"__ASSERT=="		: self.TYPE_ASSERT_EQ,
			"__ASSERT==R"		: self.TYPE_ASSERT_EQ_R,
			"__ASSERT<>"		: self.TYPE_ASSERT_NE,
			"__ASSERT>"		: self.TYPE_ASSERT_GT,
			"__ASSERT<"		: self.TYPE_ASSERT_LT,
			"__ASSERT>="		: self.TYPE_ASSERT_GE,
			"__ASSERT<="		: self.TYPE_ASSERT_LE,
			"__SLEEP"		: self.TYPE_SLEEP,
			"__STWRST"		: self.TYPE_STWRST,
			"__FEATURE"		: self.TYPE_FEATURE,
			"__GENERIC_CALL__"	: self.TYPE_GENERIC_CALL,
		}
		self.type2name_german = pivotDict(self.name2type_german)

		self.english2german = {
			"OPN"	: "AUF",
			"BEU"	: "BEA",
			"BEC"	: "BEB",
			"SF"	: "SA",
			"SD"	: "SE",
			"SP"	: "SI",
			"JU"	: "SPA",
			"JC"	: "SPB",
			"JCB"	: "SPBB",
			"JBI"	: "SPBI",
			"JNBI"	: "SPBIN",
			"JCN"	: "SPBN",
			"JNB"	: "SPBNB",
			"JL"	: "SPL",
			"JM"	: "SPM",
			"JMZ"	: "SPMZ",
			"JN"	: "SPN",
			"JO"	: "SPO",
			"JP"	: "SPP",
			"JPZ"	: "SPPZ",
			"JOS"	: "SPS",
			"JUO"	: "SPU",
			"JZ"	: "SPZ",
			"SE"	: "SV",
			"CAD"	: "TAD",
			"CAR"	: "TAR",
			"CAW"	: "TAW",
			"CDB"	: "TDB",
			"A"	: "U",
			"A("	: "U(",
			"AD"	: "UD",
			"AN"	: "UN",
			"AN("	: "UN(",
			"AW"	: "UW",
			"CD"	: "ZR",
			"CU"	: "ZV",
		}
		self.german2english = pivotDict(self.english2german)

		# Create a name2type dict for english mnemonics using the translation dict.
		self.name2type_english = {}
		for _name, _insnType in dictItems(self.name2type_german):
			with contextlib.suppress(KeyError):
				_name = self.german2english[_name]
			self.name2type_english[_name] = _insnType
		self.type2name_english = pivotDict(self.name2type_english)

AwlInsnTypes = __AwlInsnTypesClass() #+cdef-public-__AwlInsnTypesClass



# -*- coding: utf-8 -*-
#
# AWL simulator - Abstract configuration widget
#
# Copyright 2014 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.gui.util import *
from awlsim.gui.icons import *


class AbstractConfigDialog(QDialog):
	# Signal: Emitted, if any setting changed.
	settingsChanged = Signal()

	def __init__(self, project, iconName, title,
		     centralWidget, parent=None):
		QDialog.__init__(self, parent)
		self.setWindowTitle("Awlsim - " + title)
		self.setLayout(QGridLayout(self))

		self.project = project
		self.centralWidget = centralWidget

		# Add dialog header label.
		grid = QGridLayout()
		label = QLabel(self)
		label.setPixmap(getIcon(iconName).pixmap(QSize(48, 48)))
		grid.addWidget(label, 0, 0)
		label = QLabel(title, self)
		font = label.font()
		font.setPointSize(max(13, font.pointSize()))
		label.setFont(font)
		label.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
		grid.addWidget(label, 0, 1)
		# Try to center the text, regardless of the icon.
		grid.setColumnStretch(0, 1)
		grid.setColumnStretch(1, 2)
		grid.setColumnStretch(2, 1)
		self.layout().addLayout(grid, 0, 0, 1, 4)

		self.layout().addWidget(centralWidget, 1, 0, 1, 4)

		self.acceptButton = QPushButton("&Accept", self)
		self.layout().addWidget(self.acceptButton, 2, 0, 1, 3)

		self.cancelButton = QPushButton("&Cancel", self)
		self.layout().addWidget(self.cancelButton, 2, 3, 1, 1)

		self.loadFromProject()

		self.acceptButton.released.connect(self.accept)
		self.cancelButton.released.connect(self.reject)
		self.accepted.connect(self.storeToProject)

	def loadFromProject(self):
		raise NotImplementedError

	def storeToProject(self):
		raise NotImplementedError



# -*- coding: utf-8 -*-
#
# AWL simulator - Source templates
#
# Copyright 2014 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *


__all__ = [
	"Templates",
]


class Templates(object):
	# OB template
	__templateOB = """ORGANIZATION_BLOCK OB @@NR@@
	TITLE		= Insert title here
	AUTHOR		: Insert author name here
	VERSION		: 0.1
	VAR_TEMP@@TEMPVARS@@
		// ... Insert temporary variables here ...
	END_VAR
BEGIN
NETWORK
	TITLE = Insert network title here
	// ... Insert AWL/STL code here ...
	
END_ORGANIZATION_BLOCK
"""

	# OB start info variables
	__obTempVars = (
		( range(1, 1 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_SCAN_1     : BYTE;          // 1 (Cold restart scan 1 of OB @@NR@@), 3 (Scan 2-n of OB @@NR@@)
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_RESERVED_2 : BYTE;
		OB@@NR@@_PREV_CYCLE : INT;           // Cycle time of previous OB @@NR@@ scan (milliseconds)
		OB@@NR@@_MIN_CYCLE  : INT;           // Minimum cycle time of OB @@NR@@ (milliseconds)
		OB@@NR@@_MAX_CYCLE  : INT;           // Maximum cycle time of OB @@NR@@ (milliseconds)
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(10, 17 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          //Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INFO  : BYTE;          // B#16#11 (OB @@NR@@ has started)
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_RESERVED_2 : BYTE;
		OB@@NR@@_PERIOD_EXE : WORD;          // Period of execution (once, per minute/hour/day/week/month/year)
		OB@@NR@@_RESERVED_3 : INT;
		OB@@NR@@_RESERVED_4 : INT;
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(20, 23 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INF   : BYTE;          // B#16#21 (OB @@NR@@ has started)
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_RESERVED_2 : BYTE;
		OB@@NR@@_SIGN       : WORD;          // Identifier input (SIGN) attached to SRT_DINT
		OB@@NR@@_DTIME      : TIME;          // Delay time (DTIME) input to SRT_DINT instruction
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(30, 38 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INF   : BYTE;          // B#16#31 (OB @@NR@@ has started)
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_RESERVED_2 : BYTE;
		OB@@NR@@_PHS_OFFSET : INT;           // Phase offset (integer, milliseconds)
		OB@@NR@@_RESERVED_3 : INT;
		OB@@NR@@_EXC_FREQ   : INT;           // Frequency of execution (msec)
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(40, 47 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INF   : BYTE;          // B#16#41 (OB @@NR@@ has started)
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_IO_FLAG    : BYTE;          // B#16#54 (input module), B#16#55 (output module)
		OB@@NR@@_MDL_ADDR   : WORD;          // Base address of module initiating interrupt
		OB@@NR@@_POINT_ADDR : DWORD;         // Interrupt status of the module
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(55, 57 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INF   : BYTE;          // B#16#55 (OB @@NR@@ has started)
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_IO_FLAG    : BYTE;          // B#16#54 (input module), B#16#55 (output module)
		OB@@NR@@_MDL_ADDR   : WORD;          // Base address of module initiating interrupt
		OB@@NR@@_LEN        : BYTE;          // Length of information
		OB@@NR@@_TYPE       : BYTE;          // Type of alarm
		OB@@NR@@_SLOT       : BYTE;          // Slot
		OB@@NR@@_SPEC       : BYTE;          // Specifier
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(60, 60 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INF   : BYTE;          // B#16#61 (self generated), B#16#62 (external generated)
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_RESERVED_2 : BYTE;
		OB@@NR@@_JOB        : INT;           // Job sign
		OB@@NR@@_RESERVED_3 : INT;
		OB@@NR@@_RESERVED_4 : INT;
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(61, 64 + 1), """
		OB@@NR@@_EV_CLASS    : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INF    : BYTE;          // B#16#36 (OB @@NR@@ has started)
		OB@@NR@@_PRIORITY    : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR    : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1  : BYTE;
		OB@@NR@@_RESERVED_2  : BYTE;
		OB@@NR@@_GC_VIOL     : BOOL;          // Lost Global Control of DP system
		OB@@NR@@_FIRST       : BOOL;          // First Execution after startup or HALT
		OB@@NR@@_MISSED_EXEC : BYTE;          // Count of missed executions of OB @@NR@@ since it was last scheduled
		OB@@NR@@_DP_ID       : BYTE;          // DP master system ID of the synchronous DP system
		OB@@NR@@_RESERVED_3  : BYTE;
		OB@@NR@@_RESERVED_4  : WORD;
		OB@@NR@@_DATE_TIME   : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(70, 72 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Event class 7, module inserted/removed (8/9)
		OB@@NR@@_FLT_ID     : BYTE;          // Fault identifcation code
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : WORD;
		OB@@NR@@_INFO_1     : WORD;          // Info 1
		OB@@NR@@_INFO_2     : WORD;          // Info 2
		OB@@NR@@_INFO_3     : WORD;          // Info 3
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(80, 87 + 1), """
		OB@@NR@@_EV_CLASS     : BYTE;          // B#16#35, Event class 3, Entering event state, Internal fault event
		OB@@NR@@_FLT_ID       : BYTE;          // Fault identifcation code
		OB@@NR@@_PRIORITY     : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR     : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1   : BYTE;
		OB@@NR@@_RESERVED_2   : BYTE;
		OB@@NR@@_ERROR_INFO   : WORD;          // Error information on event
		OB@@NR@@_ERR_EV_CLASS : BYTE;          // Class of event causing error
		OB@@NR@@_ERR_EV_NUM   : BYTE;          // Number of event causing error
		OB@@NR@@_OB_PRIORITY  : BYTE;          // Priority of OB causing error
		OB@@NR@@_OB_NUM       : BYTE;          // Number of OB causing error
		OB@@NR@@_DATE_TIME    : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(90, 90 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // Bits 0-3 = 1 (Coming event), Bits 4-7 = 1 (Event class 1)
		OB@@NR@@_STRT_INF   : BYTE;          // 91 Cold restart, 92 Delete block, 93 Load OB @@NR@@, 95 End background cycle
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_RESERVED_2 : BYTE;
		OB@@NR@@_RESERVED_3 : INT;
		OB@@NR@@_RESERVED_4 : INT;
		OB@@NR@@_RESERVED_5 : INT;
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(100, 102 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // B#16#13, Event class 1, Entering event state, Event logged in diagnostic buffer
		OB@@NR@@_STRTUP     : BYTE;          // B#16#81/82/83/84 Method of startup
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_RESERVED_2 : BYTE;
		OB@@NR@@_STOP       : WORD;          // Event that caused CPU to stop (W#16#4XXX)
		OB@@NR@@_STRT_INFO  : DWORD;         // Information on how system started
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(121, 121 + 1), """
		OB@@NR@@_EV_CLASS   : BYTE;          // B#16#25, Event class 2, Entering event state, Internal fault event
		OB@@NR@@_SW_FLT     : BYTE;          // Software programming fault
		OB@@NR@@_PRIORITY   : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR   : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_BLK_TYPE   : BYTE;          // B#16#88/8A/8B/8C/8E Type of block fault occured in
		OB@@NR@@_RESERVED_1 : BYTE;
		OB@@NR@@_FLT_REG    : WORD;          // Specific register that caused fault
		OB@@NR@@_BLK_NUM    : WORD;          // Number of block that programming fault occured in
		OB@@NR@@_PRG_ADDR   : WORD;          // Address in block where programming fault occured
		OB@@NR@@_DATE_TIME  : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
		( range(122, 122 + 1), """
		OB@@NR@@_EV_CLASS  : BYTE;          // B#16#29, Event class 2, Entering event state, Internal fault event
		OB@@NR@@_SW_FLT    : BYTE;          // Software error code
		OB@@NR@@_PRIORITY  : BYTE;          // Priority of OB execution
		OB@@NR@@_OB_NUMBR  : BYTE;          // @@NR@@ (Organization block @@NR@@, OB @@NR@@)
		OB@@NR@@_BLK_TYPE  : BYTE;          // B#16#88/8C/8E Type of block fault occured in
		OB@@NR@@_MEM_AREA  : BYTE;          // Memory area where access error occured
		OB@@NR@@_MEM_ADDR  : WORD;          // Memory address where access error occured
		OB@@NR@@_BLK_NUM   : WORD;          // Block number in which error occured
		OB@@NR@@_PRG_ADDR  : WORD;          // Program address where error occured
		OB@@NR@@_DATE_TIME : DATE_AND_TIME; // Date and time OB @@NR@@ started"""
		),
	)


	# FC template
	__templateFC = """FUNCTION FC @@NR@@ : VOID
	TITLE		= Insert title here
	AUTHOR		: Insert author name here
	VERSION		: 0.1
	VAR_INPUT
		// Input variables
	END_VAR
	VAR_OUTPUT
		// Output variables
	END_VAR
	VAR_IN_OUT
		// In/out variables
	END_VAR
	VAR_TEMP
		// Temporary variables
	END_VAR
BEGIN
NETWORK
	TITLE = Insert network title here
	// ... Insert AWL/STL code here ...
	
END_FUNCTION
"""

	# FB template
	__templateFB = """FUNCTION_BLOCK FB @@NR@@
	TITLE		= Insert title here
	AUTHOR		: Insert author name here
	VERSION		: 0.1
	VAR
		// Static variables
	END_VAR
	VAR_INPUT
		// Input variables
	END_VAR
	VAR_OUTPUT
		// Output variables
	END_VAR
	VAR_IN_OUT
		// In/out variables
	END_VAR
	VAR_TEMP
		// Temporary variables
	END_VAR
BEGIN
NETWORK
	TITLE = Insert network title here
	// ... Insert AWL/STL code here ...
	
END_FUNCTION_BLOCK
"""

	# Instance-DB template
	__templateIDB = """DATA_BLOCK DB @@DBNR@@
	FB @@FBNR@@
	TITLE		= Insert title here
	AUTHOR		: Insert author name here
	VERSION		: 0.1
BEGIN
	// ... Insert data initializations here ...
	
END_DATA_BLOCK
"""

	# Global-DB template
	__templateGDB = """DATA_BLOCK DB @@NR@@
	TITLE		= Insert title here
	AUTHOR		: Insert author name here
	VERSION		: 0.1
	STRUCT
		// ... Insert data structure definitions here ...
		
	END_STRUCT;
BEGIN
	// ... Insert data initializations here ...
	
END_DATA_BLOCK
"""

	# UDT template
	__templateUDT = """TYPE UDT @@NR@@
	VERSION : 0.1
	STRUCT
		// ... Insert data structure definitions here ...
		
	END_STRUCT;
END_TYPE
"""

	# FC-call template
	__templateFCcall = """CALL FC @@NR@@ (
	// ... Insert parameter assignments here ...
	// VARIABLE	:= MW 0,
	// RET_VAL	:= MW 2,
)
"""

	# FB-call template
	__templateFBcall = """CALL FB @@FBNR@@, DB @@DBNR@@ (
	// ... Insert parameter assignments here ...
	// VARIABLE	:= MW 0,
)
"""

	@classmethod
	def __removeVerboseness(cls, awl):
		newAwl = []
		for line in awl.splitlines():
			stripped = line.strip()
			if stripped.startswith("TITLE") or\
			   stripped.startswith("AUTHOR") or\
			   stripped.startswith("VERSION") or\
			   stripped.startswith("NETWORK") or\
			   stripped.startswith("// ..."):
				continue
			if stripped.startswith("// Input variables") or\
			   stripped.startswith("// Output variables") or\
			   stripped.startswith("// In/out variables") or\
			   stripped.startswith("// Temporary variables") or\
			   stripped.startswith("// Static variables"):
				newAwl.append("\t\t")
				continue
			newAwl.append(line)
		return "\n".join(newAwl)

	@classmethod
	def getOB(cls, number, verbose):
		awl = cls.__templateOB[:]
		awl = awl.replace("@@NR@@", "%d" % number)
		for rnge, tempvars in cls.__obTempVars:
			if number in rnge:
				tempvars = tempvars.replace("@@NR@@", "%d" % number)
				awl = awl.replace("@@TEMPVARS@@", tempvars)
				break
		else:
			awl = awl.replace("@@TEMPVARS@@", "")
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl

	@classmethod
	def getFC(cls, number, verbose):
		awl = cls.__templateFC[:]
		awl = awl.replace("@@NR@@", "%d" % number)
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl

	@classmethod
	def getFB(cls, number, verbose):
		awl = cls.__templateFB[:]
		awl = awl.replace("@@NR@@", "%d" % number)
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl

	@classmethod
	def getInstanceDB(cls, dbNumber, fbNumber, verbose):
		awl = cls.__templateIDB[:]
		awl = awl.replace("@@DBNR@@", "%d" % dbNumber)
		awl = awl.replace("@@FBNR@@", "%d" % fbNumber)
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl

	@classmethod
	def getGlobalDB(cls, number, verbose):
		awl = cls.__templateGDB[:]
		awl = awl.replace("@@NR@@", "%d" % number)
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl

	@classmethod
	def getUDT(cls, number, verbose):
		awl = cls.__templateUDT[:]
		awl = awl.replace("@@NR@@", "%d" % number)
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl

	@classmethod
	def getFCcall(cls, number, verbose):
		awl = cls.__templateFCcall[:]
		awl = awl.replace("@@NR@@", "%d" % number)
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl

	@classmethod
	def getFBcall(cls, fbNumber, dbNumber, verbose):
		awl = cls.__templateFBcall[:]
		awl = awl.replace("@@FBNR@@", "%d" % fbNumber)
		awl = awl.replace("@@DBNR@@", "%d" % dbNumber)
		if not verbose:
			awl = cls.__removeVerboseness(awl)
		return awl



# -*- coding: utf-8 -*-
#
# AWL simulator - AWL optimizer - Remove unused labels
#
# Copyright 2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.awloptimizer.base import *

from awlsim.core.instructions.all_insns import * #+cimport

from awlsim.core.operatortypes import * #+cimport


__all__ = [
	"AwlOptimizer_LabelRemove",
]


class AwlOptimizer_LabelRemove(AwlOptimizer_Base):
	"""AWL/STL optimizer: Remove unused labels
	"""

	NAME	= "lblrem"
	STAGES	= frozenset((AwlOptimizer_Base.STAGE3, ))
	BEFORE	= frozenset()
	AFTER	= frozenset()

	def __init__(self, optimizer):
		AwlOptimizer_Base.__init__(self, optimizer)

	def run(self, insns):
		# Build a set of labels that actually are referenced.
		usedLabels = set()
		for insn in insns:
			for oper in insn.ops:
				if oper.operType == AwlOperatorTypes.LBL_REF:
					usedLabels.add(oper.immediateStr)

		# Remove unreferenced labels.
		for insn in insns:
			if insn.hasLabel() and\
			   insn.getLabel() not in usedLabels:
				insn.setLabel(None)

		return insns



# -*- coding: utf-8 -*-
#
# AWL simulator - FUP compiler - Wire
#
# Copyright 2016 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.xmlfactory import *

from awlsim.fupcompiler.base import *


class FupCompiler_WireFactory(XmlFactory):
	def parser_open(self, tag=None):
		self.inWire = False
		XmlFactory.parser_open(self, tag)

	def parser_beginTag(self, tag):
		if not self.inWire:
			if tag.name == "wire":
				self.inWire = True
				idNum = tag.getAttrInt("id")
				wire = FupCompiler_Wire(self.grid, idNum)
				if not self.grid.addWire(wire):
					raise self.Error("Invalid wire")
				return
		XmlFactory.parser_beginTag(self, tag)

	def parser_endTag(self, tag):
		if self.inWire:
			if tag.name == "wire":
				self.inWire = False
				return
		else:
			if tag.name == "wires":
				self.parser_finish()
				return
		XmlFactory.parser_endTag(self, tag)

class FupCompiler_Wire(FupCompiler_BaseObj):
	factory = FupCompiler_WireFactory

	__slots__ = (
		"grid",
		"idNum",
		"virtual",
		"connections",
	)

	def __init__(self, grid, idNum, virtual=False):
		FupCompiler_BaseObj.__init__(self)
		self.grid = grid		# FupCompiler_Grid
		self.idNum = idNum		# Wire ID
		self.virtual = virtual		# True, if this is a virtual wire

		self.clearConnections()

	def clearConnections(self):
		self.connections = set()

	def addConn(self, conn):
		self.connections.add(conn)
		conn.wire = self
		conn.wireId = self.idNum

	def removeConn(self, conn):
		self.connections.remove(conn)
		conn.wire = None
		conn.wireId = conn.WIREID_NONE



# -*- coding: utf-8 -*-
#
# AWL simulator - instructions
#
# Copyright 2012-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.exceptions import *

from awlsim.core.instructions.main import * #+cimport
from awlsim.core.operatortypes import * #+cimport
from awlsim.core.operators import * #+cimport


class AwlInsn_AbstractCall(AwlInsn): #+cdef

	__slots__ = ()

	def staticSanityChecks(self):
		if self.opCount == 1:
			# "CALL FC/SFC" or
			# "CALL #MULTIINSTANCE" or
			# "UC/CC FC/SFC/FB/SFB"
			blockOper = self.op0

			if blockOper.operType == AwlOperatorTypes.BLKREF_FC:
				try:
					codeBlock = self.cpu.fcs[blockOper.offset.byteOffset]
				except KeyError as e:
					raise AwlSimError("Called FC not found",
						rawInsn = self.rawInsn)
			elif blockOper.operType == AwlOperatorTypes.BLKREF_SFC:
				try:
					codeBlock = self.cpu.sfcs[blockOper.offset.byteOffset]
				except KeyError as e:
					raise AwlSimError("SFC %d not implemented, yet" %\
						blockOper.offset.byteOffset,
						rawInsn = self.rawInsn)
			elif blockOper.operType == AwlOperatorTypes.BLKREF_FB:
				if self.insnType == AwlInsn.TYPE_CALL:
					raise AwlSimError("Missing DB in function "
						"block call",
						rawInsn = self.rawInsn)
				try:
					codeBlock = self.cpu.fbs[blockOper.offset.byteOffset]
				except KeyError as e:
					raise AwlSimError("Called FB not found",
						rawInsn = self.rawInsn)
			elif blockOper.operType == AwlOperatorTypes.BLKREF_SFB:
				if self.insnType == AwlInsn.TYPE_CALL:
					raise AwlSimError("Missing DB in system function "
						"block call",
						rawInsn = self.rawInsn)
				try:
					codeBlock = self.cpu.sfbs[blockOper.offset.byteOffset]
				except KeyError as e:
					raise AwlSimError("SFB %d not implemented, yet" %\
						blockOper.offset.byteOffset,
						rawInsn = self.rawInsn)
			elif blockOper.operType == AwlOperatorTypes.INDIRECT:
				# Indirect call. (like UC FC[MW 0])
				codeBlock = None
			elif blockOper.operType in (AwlOperatorTypes.MULTI_FB, AwlOperatorTypes.MULTI_SFB):
				# Multi instance call (like CALL #FOO)
				if blockOper.operType == AwlOperatorTypes.MULTI_FB:
					codeBlock = self.cpu.fbs[blockOper.offset.fbNumber]
				else:
					codeBlock = self.cpu.sfbs[blockOper.offset.fbNumber]
			else:
				raise AwlSimError("Invalid CALL operand",
					rawInsn = self.rawInsn)

			if self.insnType == AwlInsn.TYPE_CALL and\
			   codeBlock and\
			   codeBlock.isFC and\
			   codeBlock.interface.interfaceFieldCount != len(self.params):
				raise AwlSimError("Call interface mismatch. "
					"Passed %d parameters, but expected %d.\n"
					"====  The block interface is:\n%s\n====" %\
					(len(self.params), codeBlock.interface.interfaceFieldCount,
					 str(codeBlock.interface)),
					rawInsn = self.rawInsn)
		elif self.opCount == 2:
			# "CALL FB/SFB"
			blockOper = self.op0
			dbOper = self.op1

			if dbOper.operType != AwlOperatorTypes.BLKREF_DB:
				raise AwlSimError("Second CALL operand is "
					"not a DB operand.",
					rawInsn = self.rawInsn)
			try:
				db = self.cpu.dbs[dbOper.offset.byteOffset]
			except KeyError as e:
				raise AwlSimError("DB used in FB call not found",
					rawInsn = self.rawInsn)
			if not db.isInstanceDB():
				raise AwlSimError("DB %d is not an instance DB" %\
					dbOper.offset.byteOffset,
					rawInsn = self.rawInsn)

			if blockOper.operType == AwlOperatorTypes.BLKREF_FB:
				try:
					fb = self.cpu.fbs[blockOper.offset.byteOffset]
				except KeyError as e:
					raise AwlSimError("Called FB not found",
						rawInsn = self.rawInsn)
				# TODO check if this is an FB-DB
				pass#TODO
			elif blockOper.operType == AwlOperatorTypes.BLKREF_SFB:
				try:
					fb = self.cpu.sfbs[blockOper.offset.byteOffset]
				except KeyError as e:
					raise AwlSimError("SFB %d not implemented, yet" %\
						blockOper.offset.byteOffset,
						rawInsn = self.rawInsn)
				# TODO check if this is an SFB-DB
				pass#TODO
			elif blockOper.operType == AwlOperatorTypes.BLKREF_FC or\
			     blockOper.operType == AwlOperatorTypes.BLKREF_SFC:
				raise AwlSimError("Calling function, but "
					"a DB was specified.",
					rawInsn = self.rawInsn)
			else:
				raise AwlSimError("Invalid CALL operand",
					rawInsn = self.rawInsn)

			if db.codeBlock.index != fb.index:
				raise AwlSimError("DB %d is not an instance DB for FB %d" %\
					(dbOper.offset.byteOffset,
					 blockOper.offset.byteOffset),
					rawInsn = self.rawInsn)
		else:
			assert(0)
		# Check parameter assignments
		for param in self.params:
			if param.isOutbound:
				if ((blockOper.operType == AwlOperatorTypes.BLKREF_FB or\
				     blockOper.operType == AwlOperatorTypes.BLKREF_SFB) and\
				    param.rvalueOp.isImmediate()) or\
				   ((blockOper.operType == AwlOperatorTypes.BLKREF_FC or\
				     blockOper.operType == AwlOperatorTypes.BLKREF_SFC) and\
				    param.rvalueOp.isImmediate() and\
				    param.rvalueOp.operType != AwlOperatorTypes.IMM_PTR):
					raise AwlSimError("Immediate value assignment '%s' "
						"to OUTPUT or IN_OUT parameter '%s' is "
						"not allowed." %\
						(str(param.rvalueOp),
						 param.lvalueName),
						rawInsn = self.rawInsn)

class AwlInsn_CALL(AwlInsn_AbstractCall): #+cdef

	__slots__ = (
		"run",
	)

	def __init__(self, cpu, rawInsn=None, **kwargs):
		AwlInsn_AbstractCall.__init__(self, cpu, AwlInsn.TYPE_CALL, rawInsn, **kwargs)
		self.assertOpCount((1, 2))

		if self.opCount == 1:			#@nocy
			self.run = self.__run_CALL_FC	#@nocy
		else:					#@nocy
			self.run = self.__run_CALL_FB	#@nocy

	def __run_CALL_FC(self): #+cdef
#@cy		cdef S7StatusWord s

		self.cpu.run_CALL(self.op0, None, self.params, False)
		s = self.cpu.statusWord
		s.OS, s.OR, s.STA, s.NER = 0, 0, 1, 0

	def __run_CALL_FB(self): #+cdef
#@cy		cdef S7StatusWord s

		self.cpu.run_CALL(self.op0, self.op1, self.params, False)
		s = self.cpu.statusWord
		s.OS, s.OR, s.STA, s.NER = 0, 0, 1, 0

#@cy	cdef run(self):
#@cy		if self.opCount == 1:
#@cy			self.__run_CALL_FC()
#@cy		else:
#@cy			self.__run_CALL_FB()

class AwlInsn_CC(AwlInsn_AbstractCall): #+cdef

	__slots__ = ()

	def __init__(self, cpu, rawInsn=None, **kwargs):
		AwlInsn_AbstractCall.__init__(self, cpu, AwlInsn.TYPE_CC, rawInsn, **kwargs)
		self.assertOpCount(1)

	def staticSanityChecks(self):
		self._warnDeprecated("Please use CALL instead.")

	def run(self): #+cdef
#@cy		cdef S7StatusWord s

		s = self.cpu.statusWord
		if s.VKE:
			self.cpu.run_CALL(self.op0, None, (), True)
		s.OS, s.OR, s.STA, s.VKE, s.NER = 0, 0, 1, 1, 0

class AwlInsn_UC(AwlInsn_AbstractCall): #+cdef

	__slots__ = ()

	def __init__(self, cpu, rawInsn=None, **kwargs):
		AwlInsn_AbstractCall.__init__(self, cpu, AwlInsn.TYPE_UC, rawInsn, **kwargs)
		self.assertOpCount(1)

	def staticSanityChecks(self):
		self._warnDeprecated("Please use CALL instead.")

	def run(self): #+cdef
#@cy		cdef S7StatusWord s

		self.cpu.run_CALL(self.op0, None, (), True)
		s = self.cpu.statusWord
		s.OS, s.OR, s.STA, s.NER = 0, 0, 1, 0



# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from media-playback-stop-4.svg

icon_stop = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
            b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
            b'A49JREFUaIHtV7uOI1UQPaeq29aQzNg9ngzxAStCUkKQSFaj'\
            b'1e6QLivET/AJfAQZEhFC/AAfgSW0KULaHUjAmgnWt4qgbt/u'\
            b'9gxi20tP1Eeyum33vVXnnno1MGPGjBkzZsyYcTQ48vkTAA0A'\
            b'PdbgcrkkAC6XS1kul3R3LhYLqeuaImInJye/b7fb3dvuV420'\
            b'3zx/8fyr3e6vr0eueyusVs03L399+e12u/1liv0B4IOnV0/8'\
            b'9etX/n/j+o9rf3r1xC8vLz8a49BYBQAAm80FfvzpB6gqVAQq'\
            b'ClUFRWGWsN/vYZZAEgAhwnyPcgUJgjBLMDN8+slnAIDz8/N6'\
            b'cgIAICL5o6AIKIQooVqDQqQ94Q6IdI4PSRAkIKIw25d9z87O'\
            b'piNwenpakreqKpCEUCDCIEMpxPaUgQqHCnRXwGxRbDRNs8AI'\
            b'jCKwWq3K6agEF1HpqcFyJYmUQoU4/aHz7vEbAIh0NpqmmU6B'\
            b'9XpdTqeqYinBHoFOhcVCsN8LUko951slPK7OO4V8UgL9+Kwq'\
            b'hXvcDwnksFKCFKgKzPxuEiPCJxRC+W/SJL64uKiTvwEAqFYA'\
            b'vPzXEdASViqKlBT7tAdxEEJxl/foSK3X60lDqL7+89WAQLjg'\
            b'AwVIKcnNnBOW5SohBBbX/cDGZASapikEKhWAhHt3luG0BpFW'\
            b'BRCJCvME9y6UIv79XhuTERhUIa3gWYGWiGYVeJgTKlGRzGJx'\
            b'IeGDQwCAuq57NWlCAqIyjOA2vvvO56pEIYTRpYuzeXF8u6vE'\
            b'JASapqm7RBRIZCYIFoco95XVGCfMCDOL53IF8o7F9AT6JU4k'\
            b'yiVyjW9VIGI8EJVep457lyBg3jnt7oMQmpTAarXqtXmW0yZK'\
            b'IYcw1Og73nZnQIOApTh9BwCHuz0Mgc1m0ynA6MLtkMYcSmzD'\
            b'ikFQRQsJ5K5tJjAzuDvMDfSj34/G94HuWzsGtAQkRgZ0s5Bm'\
            b'BVRbEjFOmBlSSiUfHiwHBmgVQJ6HeKBGeRfoh1GoMGhfBtg7'\
            b'hNComnsI/9ej68bnQX4yVoVCMpiLjsXRCrg5wLaCeDS1nMDx'\
            b'e37ODdG/fNADyrqHrEIDAt4aN5gR0pLJnY0Sp59y84r5vyNA'\
            b'AJb3eJcQOorAbvc3Hj368Gij9+H29vaoZB5L4Obx48vvX3z5'\
            b'xbPxpv4bV88+/w7AzZg1o7LIzN4H8DGA98asG4EbAD+LyG8T'\
            b'7T9jxowZM2bMmNHHP7fujLaX/o8yAAAAAElFTkSuQmCC'



__all__ = [ "find_executable", ]

import os

def find_executable(executable, path=None):
	if executable:
		import os.path
		if path is None:
			path = os.getenv("PATH", "")
		for p in path.split(":"):
			fullpath = os.path.join(p, executable)
			if os.access(fullpath, os.X_OK) and\
			   not os.path.isdir(fullpath):
				return fullpath
	return None



# -*- coding: utf-8 -*-
#
# AWL simulator - Library entry selection
#
# Copyright 2014-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.enumeration import *
from awlsim.common.wordpacker import *
from awlsim.common.xmlfactory import *
from awlsim.common.util import *

import hashlib


class AwlLibEntrySelectionFactory(XmlFactory):
	def parser_open(self, tag=None):
		libSel = self.libSel
		if tag:
			name = tag.getAttr("name")
			entryType = tag.getAttr("type")
			index = tag.getAttrInt("index")
			effectiveIndex = tag.getAttrInt("effective_index", index)
			libSel.setLibName(name)
			libSel.setEntryType(entryType)
			libSel.setEntryIndex(index)
			libSel.setEffectiveEntryIndex(effectiveIndex)
		XmlFactory.parser_open(self, tag)

	def parser_beginTag(self, tag):
		XmlFactory.parser_beginTag(self, tag)

	def parser_endTag(self, tag):
		if tag.name == "lib_selection":
			self.parser_finish()
			return
		XmlFactory.parser_endTag(self, tag)

	def composer_getTags(self):
		project, libSel = self.project, self.libSel

		childTags = []

		tags = [self.Tag(name="lib_selection",
				 comment="\nStandard library selection",
				 attrs={
					"name"		  : str(libSel.getLibName()),
					"type"		  : str(libSel.getEntryTypeStr()),
					"index"		  : str(int(libSel.getEntryIndex())),
					"effective_index" : str(int(libSel.getEffectiveEntryIndex())),
				 },
				 tags=childTags),
		]
		return tags

class AwlLibEntrySelection(object):
	"""AWL library entry selection."""

	IDENT_HASH	= hashlib.sha256

	# Library entry type.
	# This enumeration is awlsim-coreserver API!
	EnumGen.start
	TYPE_UNKNOWN	= EnumGen.item
	TYPE_FC		= EnumGen.item
	TYPE_FB		= EnumGen.item
	EnumGen.end

	factory = AwlLibEntrySelectionFactory

	def __init__(self, libName="",
		     entryType=TYPE_UNKNOWN,
		     entryIndex=-1, effectiveEntryIndex=-1):
		self.__identHash = None
		self.setLibName(libName)
		self.setEntryType(entryType)
		self.setEntryIndex(entryIndex)
		self.setEffectiveEntryIndex(effectiveEntryIndex)

	def setLibName(self, libName):
		self.__libName = libName.strip()
		self.__identHash = None

	def getLibName(self):
		return self.__libName

	def setEntryType(self, entryType):
		if isString(entryType):
			if entryType.upper() == "FC":
				entryType = self.TYPE_FC
			elif entryType.upper() == "FB":
				entryType = self.TYPE_FB
		if entryType not in (self.TYPE_UNKNOWN,
				     self.TYPE_FC,
				     self.TYPE_FB):
			raise AwlSimError("Library selection: "
				"Invalid entry type %d." %\
				entryType)
		self.__entryType = entryType
		self.__identHash = None

	def getEntryType(self):
		return self.__entryType

	def getEntryTypeStr(self):
		return {
			self.TYPE_UNKNOWN	: "UNKNOWN",
			self.TYPE_FC		: "FC",
			self.TYPE_FB		: "FB",
		}[self.getEntryType()]

	def setEntryIndex(self, entryIndex):
		if entryIndex < -1 or entryIndex > 0xFFFF:
			raise AwlSimError("Library selection: "
				"Invalid entry index %d." %\
				entryIndex)
		self.__entryIndex = entryIndex
		self.__identHash = None

	def getEntryIndex(self):
		return self.__entryIndex

	def setEffectiveEntryIndex(self, effectiveEntryIndex):
		if effectiveEntryIndex < -1 or effectiveEntryIndex > 0xFFFF:
			raise AwlSimError("Library selection: "
				"Invalid effective entry index %d." %\
				effectiveEntryIndex)
		self.__effectiveEntryIndex = effectiveEntryIndex
		self.__identHash = None

	def getEffectiveEntryIndex(self):
		return self.__effectiveEntryIndex

	def isValid(self):
		return self.__libName and\
		       self.__entryType in (self.TYPE_FC, self.TYPE_FB) and\
		       self.__entryIndex >= 0 and\
		       self.__entryIndex <= 0xFFFF and\
		       self.__effectiveEntryIndex >= 0 and\
		       self.__effectiveEntryIndex <= 0xFFFF

	def getIdentHash(self):
		"""Get the unique identification hash for this
		library selection descriptor.
		"""
		if not self.__identHash:
			# Calculate the ident hash
			h = self.IDENT_HASH(b"AwlLibEntrySelection")
			h.update(self.__libName.encode("utf-8", "ignore"))
			h.update(WordPacker.toBytes(bytearray(4), 32, 0,
						    self.__entryIndex))
			h.update(WordPacker.toBytes(bytearray(4), 32, 0,
						    self.__effectiveEntryIndex))
			self.__identHash = h.digest()
		return self.__identHash

	def getIdentHashStr(self):
		return bytesToHexStr(self.getIdentHash())

	def __eq__(self, other):
		return self.getIdentHash() == other.getIdentHash()

	def __ne__(self, other):
		return not self.__eq__(other)

	def __repr__(self):
		asStr = ""
		if self.getEntryIndex() != self.getEffectiveEntryIndex():
			asStr = " (as %s %d)" % (self.getEntryTypeStr(),
						 self.getEffectiveEntryIndex())
		return "%s: %s %d%s" % (
			self.getLibName(),
			self.getEntryTypeStr(),
			self.getEntryIndex(),
			asStr,
		)



# -*- coding: utf-8 -*-
#
# AWL simulator - Abstract hardware interface
#
# Copyright 2013-2017 Michael Buesch <m@bues.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *

from awlsim.common.exceptions import *

from awlsim.core.offset import * #+cimport
from awlsim.core.hardware_loader import *
from awlsim.core.hardware_params import *
#from awlsim.core.hardware cimport * #@cy


__all__ = [ "AbstractHardwareInterface", ]


class AbstractHardwareInterface(object): #+cdef
	"""Abstract hardware interface class.
	This class must be subclassed in the hardware interface module.
	The subclass must be named 'HardwareInterface' for the automatic loading
	mechanism to work."""

	# The name of the module. Overload in the subclass.
	name = "<unnamed>"

	# The parameter descriptors.
	paramDescs = []
	# The standard parameters.
	__standardParamDescs = [
		HwParamDesc_bool("removeOnReset",
				 defaultValue = True,
				 description = "If set to 'True' the module will "
					       "be removed on CPU reset."),
		HwParamDesc_int("inputAddressBase",
				defaultValue = 0, minValue = 0,
				description = "Start address in input address range"),
		HwParamDesc_int("outputAddressBase",
				defaultValue = 0, minValue = 0,
				description = "Start address in output address range"),
	]

	@classmethod
	def getParamDescs(cls):
		"""Get all parameter descriptors for this class."""
		descs = cls.__standardParamDescs[:]
		descs.extend(cls.paramDescs)
		return descs

	@classmethod
	def getParamDesc(cls, paramName):
		"""Get one parameter descriptor."""
		for desc in cls.getParamDescs():
			if desc.match(paramName):
				return desc
		return None

	@classmethod
	def getModuleInfo(cls):
		"""Get module information. Returns a string."""
		ret = []
		ret.append("Hardware module '%s':" % cls.name)
		ret.append("")
		ret.append("Parameters:")
		for desc in cls.getParamDescs():
			if desc.mandatory:
				defStr = "mandatory"
			else:
				defStr = "default: %s" % str(getattr(desc, "defaultValue", None))
			ret.append(" %s = %s (%s)%s" %\
				(desc.name, desc.typeStr.upper(),
				 defStr,
				 ("  -  " + desc.description) if desc.description else ""))
		return "\n".join(ret)

	def __init__(self, sim, parameters={}):
		"""Constructs the abstract hardware interface.
		'sim' is the AwlSim instance.
		'parameters' is a dict of hardware specific parameters."""
		self.sim = sim
		self.__running = False
		self.__parseParameters(parameters)

		# Get the base addresses for convenience.
		self.inputAddressBase = self.getParamValueByName("inputAddressBase")
		self.outputAddressBase = self.getParamValueByName("outputAddressBase")

	def startup(self):
		"""Initialize access to the hardware."""
		if not self.__running:
			self.doStartup()
			self.__running = True

	def doStartup(self):
		"""Actually initialize access to hardware.
		Overload this method, if the hardware needs initialization"""
		pass

	def shutdown(self):
		"""Shutdown access to the hardware."""
		if self.__running:
			self.__running = False
			self.doShutdown()

	def doShutdown(self):
		"""Actually shutdown access to hardware.
		Overload this method, if the hardware needs initialization"""
		pass

	def readInputs(self): #+cdef
		"""Read all hardware input data and store it in the PAE.
		The implementation is supposed to put the data directly
		into the cpu memory.
		Overload this method, if the hardware has inputs."""
		pass

	def writeOutputs(self): #+cdef
		"""Write all hardware output data (read from PAA).
		The implementation is supposed to read the data directly
		from the cpu memory.
		Overload this method, if the hardware has outputs."""
		pass

	def directReadInput(self, accessWidth, accessOffset): #@nocy
#@cy	cdef bytearray directReadInput(self, uint32_t accessWidth, uint32_t accessOffset):
		"""Direct peripheral input data read operation.
		'accessWidth' is the width of the access, in bits.
		'accessOffset' is the byte offset of the access.
		The read data is returned.
		An empty bytearray is returned if the 'accessOffset' is not in
		this hardware's range.
		Overload this method, if the hardware has inputs and
		supports direct peripheral access."""
		return bytearray()

	def directWriteOutput(self, accessWidth, accessOffset, data): #@nocy
#@cy	cdef ExBool_t directWriteOutput(self, uint32_t accessWidth, uint32_t accessOffset, bytearray data) except ExBool_val:
		"""Direct peripheral output data write operation.
		'accessWidth' is the width of the access, in bits.
		'accessOffset' is the byte offset of the access.
		'data' is the data to write.
		True is returned, if the hardware successfully stored the value.
		False is returned, if the 'accessOffset' is not in this
		hardware's range.
		Overload this method, if the hardware has outputs and
		supports direct peripheral access."""
		return False

	def __repr__(self):
		return "HardwareInterface: %s" % self.name

	def raiseException(self, errorText):
		"""Throw an exception."""
		raise AwlSimError("['%s' hardware module] %s" %\
				  (self.name, errorText))

	def paramErrorHandler(self, name, errorText):
		"""Default parameter error handler."""
		self.raiseException("Parameter '%s': %s" % (name, errorText))

	def __parseParameters(self, parameters):
		# Parse the parameters.
		self.__paramsByName = {}
		self.__paramsByDescType = {}
		for name, value in dictItems(parameters):
			for desc in self.getParamDescs():
				if desc.match(name):
					break
			else:
				self.paramErrorHandler(name,
					"Invalid parameter. The parameter '%s' is "
					"unknown to the '%s' hardware module." %\
					(name, self.name))
			try:
				parsedValue = desc.parse(value)
			except HwParamDesc.ParseError as e:
				self.paramErrorHandler(name, str(e))
			self.__paramsByName[name] = parsedValue
			self.__paramsByDescType.setdefault(type(desc), []).append(
					(name, parsedValue))

		# Check mandatory parameters
		for desc in self.getParamDescs():
			if not desc.mandatory:
				continue
			if desc.name not in dictKeys(self.__paramsByName):
				self.paramErrorHandler(desc.name,
					"Mandatory parameter not specified")

	def getParamValueByName(self, name):
		"""This is the main method to get a parameter value by name
		from the hardware module.
		'name' is the name string of the parameter.
		"""

		descs = [ d for d in self.getParamDescs() if d.match(name) ]
		# Programming error, if getParamValueByName() was called with a name
		# that was not declared in paramDescs.
		assert(descs)

		# Get the value.
		try:
			return self.__paramsByName[name]
		except KeyError:
			return getattr(descs[0], "defaultValue", None)

	getParam = getParamValueByName # Old deprecated getParam() API

	def getParamsByDescType(self, descType):
		"""This is the main method to get parameter name/value pairs
		for a given param desc type.
		This returns a list of (name, value) tuples.
		"""
		try:
			return self.__paramsByDescType[descType]
		except KeyError:
			return []
